/*
 *  Copyright (c) xlightweb.org, 2008 - 2009. All rights reserved.
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please refer to the LGPL license at: http://www.gnu.org/copyleft/lesser.txt
 * The latest copy of this software may be found on http://www.xlightweb.org/
 */
package org.xlightweb;

import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.util.Enumeration;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;

import org.xsocket.DataConverter;






/**
 * Implementation base of a message
 * 
 * @author grro@xlightweb.org
 */
abstract class AbstractHttpMessage implements IHttpMessage {
	
	private final AtomicReference<NonBlockingBodyDataSource> bodyDataSourceRef = new AtomicReference<NonBlockingBodyDataSource>();

	

	/**
	 * {@inheritDoc}
	 */
	public abstract IHttpHeader getMessageHeader();
	

	/**
	 * {@inheritDoc}
	 */
	public final void setAttribute(String name, Object o) {
		getMessageHeader().setAttribute(name, o);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final Object getAttribute(String name) {
		return getMessageHeader().getAttribute(name);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public final Enumeration getAttributeNames() {
		return getMessageHeader().getAttributeNames();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final Set<String> getAttributeNameSet() {
		return getMessageHeader().getAttributeNameSet();
	}

	

    
    static NonBlockingBodyDataSource newDataSource(ByteBuffer[] body, String encoding, boolean compress) throws IOException, UnsupportedEncodingException {
        
        if (compress) {
            byte[] data = HttpUtils.compress(DataConverter.toBytes(body));
            return new InMemoryBodyDataSource(encoding, HttpUtils.newMultimodeExecutor(), data);
        } else {
            return new InMemoryBodyDataSource(encoding, HttpUtils.newMultimodeExecutor(), body);
        }
    }

    

	

    
    final void enhanceHeaderWithBodySize(NonBlockingBodyDataSource body) throws IOException {
        
        if (AbstractHttpConnection.isChunkedTransferEncoding(getMessageHeader())) {
            return;
        } else if (getMessageHeader().getContentLength() >= 0) {
            return;
        }
        
        if (body.isComplete()) {
            getMessageHeader().setContentLength(body.available());
            
        } else {
                
            if (body instanceof FullMessageBodyDataSource) {
                getMessageHeader().setContentLength(((FullMessageBodyDataSource) body).getSize());
            
            } else if (body instanceof FullMessageChunkedBodyDataSource) {
                getMessageHeader().setTransferEncoding("chunked");
                
            } else if (body instanceof MultipartByteRangeMessageBodyDataSource) {
                if (getMessageHeader().getContentType().toLowerCase().startsWith("multipart/byteranges")) {
                    return;
                }
                String separator = ((MultipartByteRangeMessageBodyDataSource) body).getSeparator();
                getMessageHeader().setContentType("Content-Type: multipart/byteranges; boundary=" + separator); 
                
            } 
        }        
    }
	
	
	/**
	 * sets the body data source 
	 * 
	 * @param body  the body data source 
	 */
	final void setBodyDataSource(NonBlockingBodyDataSource body) throws IOException {
        if (body != null) {
            bodyDataSourceRef.set(body);
        } 
	}
	
	



	

	/**
	 * {@inheritDoc}
	 */
	public final NonBlockingBodyDataSource getNonBlockingBody() throws IOException {
	    NonBlockingBodyDataSource ds = bodyDataSourceRef.get();
	    if (ds == null)  {
	        try {
	            ds = new InMemoryBodyDataSource(DEFAULT_ENCODING, HttpUtils.newMultimodeExecutor());
	            ds.setComplete();
	        } catch (IOException ioe) {
	            throw new RuntimeException(ioe);
	        }
	    } 
	        
	    return ds;
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final BlockingBodyDataSource getBlockingBody() throws IOException {
		return new BlockingBodyDataSource(getNonBlockingBody());
	}

	
	/**
	 * returns true if the message has a body
	 * 
	 * @return  true, if the message has a body
	 */
	public final boolean hasBody() {
		return bodyDataSourceRef.get() != null;
	}
	

	/**
	 * {@inheritDoc}
	 */
	public void addHeaderLine(String line) {
		getMessageHeader().addHeaderLine(line);
	}
	
	
	/**
     * {@inheritDoc}
     */
	public void addHeaderlines(String... lines) {
	    getMessageHeader().addHeaderlines(lines);
	    
	}
	

	/**
	 * {@inheritDoc}
	 */
	public final void addHeader(String headername, String headervalue) {
		getMessageHeader().addHeader(headername, headervalue);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final void setHeader(String headername, String headervalue) {
		getMessageHeader().setHeader(headername, headervalue);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public final void removeHeader(String headername) {
		getMessageHeader().removeHeader(headername);
	}
	
	
	
	/**
	 * {@inheritDoc}
	 */
	public boolean containsHeader(String headername) {
		return getMessageHeader().containsHeader(headername);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public final Enumeration getHeaderNames() {
		return getMessageHeader().getHeaderNames();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final Set<String> getHeaderNameSet() {
		return getMessageHeader().getHeaderNameSet();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final String getHeader(String headername) {
		return getMessageHeader().getHeader(headername);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public final List<String> getHeaderList(String headername) {
		return getMessageHeader().getHeaderList(headername);
	}
	
	/**
	 * {@inheritDoc}
	 */
	@SuppressWarnings("unchecked")
	public final Enumeration getHeaders(String headername) {
		return getMessageHeader().getHeaders(headername);
	}

	
	/**
	 * {@inheritDoc}
	 */
	public void setContentType(String type) {
		getMessageHeader().setContentType(type);
	}

	/**
	 * {@inheritDoc}
	 */
	public String getContentType() {
		return getMessageHeader().getContentType();
	}

	

	/**
	 * {@inheritDoc}
	 */
	public String getCharacterEncoding() {
		return getMessageHeader().getCharacterEncoding();
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public void setContentLength(int length) {
		getMessageHeader().setContentLength(length);
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public int getContentLength() {
		return getMessageHeader().getContentLength();
	}

	
    /**
     * {@inheritDoc}
     */
	public String getDisposition() {
	    return getMessageHeader().getDisposition();
	}
	
	
	/**
     * {@inheritDoc}
     */
	public String getDispositionType() {
	    return getMessageHeader().getDispositionType();
	}
	
	
	/**
     * {@inheritDoc}
     */
	public String getDispositionParam(String name) {
	    return getMessageHeader().getDispositionParam(name);
	}
	

	
	/**
	 * {@inheritDoc}
	 */
	public void setTransferEncoding(String transferEncoding) {
		getMessageHeader().setTransferEncoding(transferEncoding);
		removeHeader("Content-length");
	}
	
	
	/**
	 * {@inheritDoc}
	 */
	public String getTransferEncoding() {
		return getMessageHeader().getTransferEncoding();
	}
	
	
	public void removeHopByHopHeaders() {
	    getMessageHeader().removeHopByHopHeaders();
	}

	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		if (bodyDataSourceRef.get() == null) {
			return getMessageHeader().toString() + "\r\n";
		} else {
		    return getMessageHeader().toString() + "\r\n" + bodyDataSourceRef.toString();
		}
	}
}